package se.tada.preproc;

import java.util.HashMap;
import java.io.Writer;
import java.io.IOException;

import se.tada.util.eval.EvalException;
import se.tada.util.eval.Expression;

class DefineDirective extends Unit implements Comparable
{
	private int			m_paramCount;
	private String		m_name;
	private Expression	m_expression;

	public DefineDirective(PreProcessor owner, String fileName, int startLine)
	{
		super(owner, fileName, startLine);
		m_paramCount = Integer.MIN_VALUE;
		m_name = "";
		m_expression = null;
	}

	public Number toNumber()
	throws ParseException
	{
		if(m_contents == null || m_contents.length() == 0
				|| m_paramCount != Integer.MIN_VALUE)
			return new Integer(0);

		if(m_expression == null)
		{
			m_expression = m_owner.compile(m_contents);
			if(m_expression == null)
				return new Integer(0);
		}
		try
		{
			Object num = m_expression.eval(m_owner.getMacros());
			if(!(num instanceof Number))
				throw new ParseException("'" + m_contents + "' does not evalutate to a number");
			return (Number)num;
		}
		catch(EvalException e)
		{
			throw new ParseException(e);
		}
	}

	public boolean equals(Object anObject)
	{
		// Uncomparable yields not equal.
		//
		try
		{
			return this.compareTo(anObject) == 0;
		}
		catch(RuntimeException e)
		{}
		return false;
	}

	public int compareTo(Object v)
	{
		if(!(v instanceof Number))
			throw new IllegalArgumentException("not a number");

		try
		{
			double d1 = this.toNumber().doubleValue();
			double d2 = ((Number)v).doubleValue();
			return (d1 > d2) ? 1 : ((d1 == d2) ? 0 : -1);
		}
		catch(ParseException e)
		{
			throw new RuntimeException(e.getMessage());
		}
	}

	public void expand()
	throws ParseException
	{
		m_owner.defineMacro(this);
		super.expand();
	}

	public int parse(char[] buf, int bIdx)
	{
		Integer pInt;
		char c;
		String argName;
		boolean inQuote = false;
		HashMap params = null;
		int paramNo = 0;
		int[] lcnt = new int[]{bIdx};
		String tmp = this.cleanDirectiveLine(buf, lcnt);
		int rIdx = lcnt[0]; // Index to be returned

		buf = tmp.toCharArray();
		int bTop = buf.length;
		bIdx = 0;

		boolean depCheck = m_owner.doingDependencyCheck();

		if(bTop == 0 || !(Character.isLetter(c = buf[0]) || c == '_'))
		{
			this.parseError("bad define");
			return rIdx;
		}

		lcnt[0] = bIdx;
		m_name = getIdentifier(buf, lcnt);
		bIdx = lcnt[0];
		c = (bIdx < bTop ? buf[bIdx] : 0);

		if(c == '(')
		{
			// Parse macro parameter list.
			//
			if(!depCheck)
				params = new HashMap();

			++bIdx;
			while(bIdx < bTop)
			{
				bIdx = skipWhite(buf, bIdx, lcnt);
				m_lineCount += lcnt[0];

				if(bIdx >= bTop)
					break;

				if((c = buf[bIdx]) == ')')
					break;

				if(++paramNo > 1)
				{
					if(c != ',')
					{
						if(!depCheck)
						{
							bIdx = skipToStartOfLine(buf, bIdx, lcnt);
							m_lineCount += lcnt[0];
							this.parseError("bad parameter list : "
									+ String.valueOf(buf, 0, bIdx));
						}
						return rIdx;
					}

					bIdx = skipWhite(buf, bIdx, lcnt);
					m_lineCount += lcnt[0];
					if(bIdx >= bTop)
						break;
					c = buf[bIdx];
				}

				if(!(Character.isLetter(c) || c == '_'))
				{
					if(!depCheck)
					{
						bIdx = skipToStartOfLine(buf, bIdx, lcnt);
						m_lineCount += lcnt[0];
						this.parseError("bad parameter list : "
								+ String.valueOf(buf, 0, bIdx));
					}
					return rIdx;
				}

				if(!depCheck)
				{
					lcnt[0] = bIdx;
					argName = getIdentifier(buf, lcnt);
					bIdx = lcnt[0];
					params.put(argName, new Integer(paramNo));
				}
			}

			if(c != ')')
			{
				this.parseError("unterminated param list");
				return rIdx;
			}
			bIdx = skipWhite(buf, ++bIdx, lcnt);
			m_lineCount += lcnt[0];
			m_paramCount = paramNo;
		}
		else
		{
			bIdx = skipWhite(buf, bIdx, lcnt);
			m_lineCount += lcnt[0];
		}

		// Everything following this point should be expanded. Let's
		// parse the rest, substituting parameter occurences.
		//

		// First of all, we need our own buffer.
		//
		int oIdx = 0;
		while(bIdx < bTop)
		{
			c = buf[bIdx++];
			if(inQuote)
			{
				// We don't recognize macros within quotes
				//
				buf[oIdx++] = c;
				if(c == '\\')
				{
					if(bIdx < bTop)
						buf[oIdx++] = buf[bIdx++];
					continue;
				}
				if(c == '"')
					inQuote = false;
				continue;
			}

			switch(c)
			{
				case ESCNL_MARKER:
					++m_lineCount;

				// Fall through to whitespace
				case ' ':
				case '\f':
				case '\t':
				case NULL_MARKER:
					// Within a macro, we simply replace all sequences
					// of whitespace with a single space.
					//
					buf[oIdx++] = ' ';
					if(bIdx < bTop)
					{
						bIdx = skipWhite(buf, bIdx, lcnt);
						m_lineCount += lcnt[0];
					}
					break;

				case '\\':
					buf[oIdx++] = c;
					if(bIdx < bTop)
						buf[oIdx++] = buf[bIdx++];
					break;

				case '"':
					inQuote = true;
					buf[oIdx++] = c;
					break;

				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':

					// Numbers must be skipped in order to avoid
					// attempt to do parameter substitusions on things
					// line 012ADAM
					//
					buf[oIdx++] = c;
					while(bIdx < bTop
							&& (Character.isLetterOrDigit(c = buf[bIdx]) || c == '_'))
					{
						buf[oIdx++] = c;
						++bIdx;
					}
					break;

				default:
					if(!Character.isLetter(c))
					{
						buf[oIdx++] = c;
						break;
					}

				// Fall through to '_':
				//
				case '_':
					// Possible parameter substitution
					//
					if(!depCheck && paramNo > 0)
					{
						int mIdx = --bIdx;
						lcnt[0] = bIdx;
						argName = getIdentifier(buf, lcnt);
						bIdx = lcnt[0];

						if((pInt = (Integer)params.get(argName)) == null)
						{
							// No such parameter. Copy verbatim
							//
							while(mIdx < bIdx)
								buf[oIdx++] = buf[mIdx++];
							break;
						}

						// Mark next byte as parameter number
						//
						buf[oIdx++] = (char)PARAM_MARKER;

						// And print the parameter number.
						//
						c = (char)pInt.intValue();
					}
					buf[oIdx++] = c;
					break;
			}
		}

		if(depCheck && paramNo > 0)
			//
			// if directives will never cause an expansion
			//
			m_contents = null;
		else
			m_contents = String.valueOf(buf, 0, oIdx);
		return rIdx;
	}

	public void expandFinal(Writer here, Unit owr, String aContents)
			throws IOException
	{
		PlainText tp = new PlainText(m_owner, owr.getFileName(), owr
				.getStartLine());
		tp.setContents(aContents);
		tp.expandTo(here);
	}

	public int expandMacro(Writer here, Unit owr, char[] buf, int bIdx,
			int[] eatNl) throws IOException
	{
		// Parse parameter list if we have parameters.
		//
		if(m_contents == null)
			//
			// Expansion of invalid macro.
			//
			return bIdx;

		StringBuffer mark = new StringBuffer(40);
		mark.append((char)SOURCE_PUSH);
		mark.append(this.getFileName());
		mark.append((char)SOURCE_PUSH);
		mark.append(this.getStartLine() + 1);
		mark.append((char)SOURCE_PUSH);
		here.write(mark.toString());
		int ret = (m_paramCount == Integer.MIN_VALUE) ? expandNoParams(here,
				owr, buf, bIdx) : expandWithParams(here, owr, buf, bIdx, eatNl);
		here.write(SOURCE_POP);
		return ret;
	}

	public int expandNoParams(Writer here, Unit owr, char[] buf, int bIdx)
			throws IOException
	{
		this.expandFinal(here, owr, m_contents);
		return bIdx;
	}

	public int expandWithParams(Writer here, Unit owr, char[] buf, int bIdx,
			int[] eatNl) throws IOException
	{
		char c = 0;
		StringBuffer out = new StringBuffer();
		String[] params = new String[m_paramCount];
		int depth = 1;
		boolean inQuote = false;
		int pcnt = 0;
		int wcnt = 0;
		int[] lcnt = new int[1];
		int bTop = buf.length;
		int start;
		int rIdx;

		// Parse the parameters. Parameters are delimited by ','. The
		// delimiter should not be recognized if delimited by quotes
		// "x" or ',' or within parenthesis. Comments have already
		// been stripped from input.
		//
		bIdx = skipAllWhite(buf, bIdx, lcnt);
		wcnt = lcnt[0];
		if(bIdx >= bTop - 1 || buf[bIdx] != '(')
		{
			owr.parseError("bad parameter list");
			return bIdx;
		}

		++bIdx;
		for(;;)
		{
			bIdx = skipAllWhite(buf, bIdx, lcnt);
			wcnt += lcnt[0];
			start = bIdx;
			while(bIdx < bTop)
			{
				c = buf[bIdx++];
				if(inQuote)
				{
					if(c == '\\')
					{
						if(bIdx < bTop)
							++bIdx;
						continue;
					}
					if(c == '"')
						inQuote = false;
					else if(c == '\n' || c == ESCNL_MARKER)
						++wcnt;
					continue;
				}

				if(depth > 1)
				{
					switch(c)
					{
						case '(':
							++depth;
							continue;
						case ')':
							--depth;
							continue;
						case ESCNL_MARKER:
						case '\n':
							lcnt[0]++;
					}
					continue;
				}

				switch(c)
				{
					case '(':
						depth++;
						continue;

					case '\\':
						if(bIdx < bTop)
							++bIdx;
						continue;

					case ESCNL_MARKER:
					case '\n':
						++wcnt;
						continue;

					case '\'':
						bIdx = skipSingleQuote(buf, bIdx);
						continue;

					case '\"':
						inQuote = true;
						continue;

					case ')':
					case ',':
						break;

					default:
						continue;
				}
				break;
			}

			if(++pcnt > m_paramCount)
				break;

			if(c != ',')
				break;

			int cIdx = bIdx - 2;
			while(cIdx > start && Character.isWhitespace(buf[cIdx]))
				--cIdx;
			++cIdx;
			params[pcnt] = String.valueOf(buf, start, cIdx - start);
		}

		--bIdx;
		eatNl[0] += wcnt;

		if(pcnt != m_paramCount)
		{
			if(c == ')' && m_paramCount == 0 && pcnt == 1)
			{
				// Allow one single whitespace argument.
				//
				while(start < bIdx)
				{
					c = buf[start];
					if(Character.isWhitespace(c) || c == NULL_MARKER
							|| c == ESCNL_MARKER)
					{
						++start;
						continue;
					}
				}
				if(start < bIdx)
				{
					owr.parseError("incorrect number of arguments");
					return bIdx;
				}
				pcnt = 0;
			}
			else
			{
				owr.parseError("incorrect number of arguments");
				return bIdx;
			}
		}

		if(c != ')')
		{
			owr.parseError("unterminated macro list");
			return bIdx;
		}

		if(pcnt > 0)
		{
			int cIdx = bIdx - 1;
			while(cIdx > start)
			{
				c = buf[cIdx];
				if(!(Character.isWhitespace(c) || c == NULL_MARKER || c == ESCNL_MARKER))
					break;
				--cIdx;
			}
			++cIdx;
			params[pcnt] = String.valueOf(buf, start, cIdx - start);
		}
		rIdx = ++bIdx;

		// Now, substitute parameters
		//
		buf = m_contents.toCharArray();
		bIdx = 0;
		bTop = buf.length;
		start = 0;

		while(bIdx < bTop)
		{
			int lIdx = bIdx;
			bIdx = skipWhite(buf, bIdx, lcnt);
			switch(bIdx < bTop ? buf[bIdx] : 0)
			{
				case '#':
					switch(bIdx + 1 < bTop ? buf[bIdx + 1] : 0)
					{
						case '#':
							// Concatenate. 'lIdx' points to the last non
							// whitespace
							// character before '##', skip to first non
							// whitespace
							// after '##' and continue.
							//
							bIdx = skipWhite(buf, bIdx + 2, lcnt);
							out.append(buf, start, lIdx - start);
							start = bIdx;
							continue;

						case PARAM_MARKER:
							// Stringify if the next token is a parameter. If
							// not
							// this is an error.
							//
							out.append(buf, start, bIdx - start);
							bIdx += 2;
							out.append('"');
							out.append(params[buf[bIdx++]]);
							out.append('"');
							start = bIdx;
							continue;
					}
					owr.parseError("# must be followed by parameter");
					return bIdx;

				case PARAM_MARKER:
					if(bIdx > start)
						out.append(buf, start, bIdx - start);
					++bIdx;
					out.append(params[buf[bIdx++]]);
					start = bIdx;
					continue;

				default:
					++bIdx;
					continue;
			}
		}
		if(bIdx > start)
			out.append(buf, start, bIdx - start);
		this.expandFinal(here, owr, out.toString());
		return rIdx;
	}

	public String name()
	{
		return m_name;
	}
}
