package se.tada.preproc;

import java.util.Vector;

import se.tada.util.eval.EvalException;
import se.tada.util.eval.Expression;

class IfDirective extends Unit
{
	public IfDirective(PreProcessor owner, String fileName, int startLine)
	{
		super(owner, fileName, startLine);
		m_trueBlock = new Unit[0];
		m_falseBlock = m_trueBlock;
		m_expression = null;
		m_totalLineCount = 0;
	}

	public static boolean booleanOf(Object v)
	{
		if(v instanceof Boolean)
			return ((Boolean)v).booleanValue();

		if(v instanceof Integer)
			return ((Integer)v).intValue() != 0;

		if(v instanceof Double)
			return ((Double)v).doubleValue() != 0.0;

		if(v instanceof Character)
			return ((Character)v).charValue() != 0;

		if(v instanceof Long)
			return ((Long)v).longValue() != 0L;

		if(v instanceof Short)
			return ((Short)v).shortValue() != 0;

		if(v instanceof Byte)
			return ((Byte)v).byteValue() != 0;

		if(v instanceof Float)
			return ((Float)v).floatValue() != 0.0;

		return v != null;
	}

	public void expand()
	throws ParseException
	{
		if(m_expression == null)
		{
			super.expand();
			return;
		}

		int top;
		int idx;
		Unit[] currentBlock;
		int cLine = m_startLine + m_lineCount;

		try
		{
			if(m_expression.isTrue(m_owner.getMacros()))
				currentBlock = m_trueBlock;
			else
			{
				// We need to create a line mark since we now skip the
				// true block.
				//
				top = m_trueBlock.length;
				for(idx = 0; idx < top; ++idx)
					cLine += m_trueBlock[idx].getLineCount();
				cLine++; // The #else or #elif
				currentBlock = m_falseBlock;
			}
		}
		catch(EvalException e)
		{
			throw new ParseException(e);
		}

		m_owner.setLine(cLine);
		if((top = currentBlock.length) > 0)
		{
			for(idx = 0; idx < top; ++idx)
				currentBlock[idx].expand();
		}
		m_owner.setLine(m_startLine + m_totalLineCount);
	}

	public int getLineCount()
	{
		return m_totalLineCount;
	}

	public int parse(char[] buf, int bIdx)
	throws ParseException
	{
		bIdx = this.parseExpression(buf, bIdx);
		return this.parseContents(buf, bIdx);
	}

	public int parseExpression(char[] buf, int bIdx)
	throws ParseException
	{
		// Scan until next nonescaped newline.
		//
		Expression ecb;
		int[] bpt = new int[]{bIdx};
		String expr = this.cleanDirectiveLine(buf, bpt);
		if((ecb = m_owner.compile(expr)) == null)
			this.parseError("bad #if expression : " + expr);
		else
		{
			m_expression = ecb;
			m_contents = expr;
		}
		return bpt[0];
	}

	public int parseContents(char[] buf, int bIdx)
	throws ParseException
	{
		// Parse the if part of the expression. It ends when we find one of
		// #else, #elif and #endif
		//
		Vector currentBlock;
		boolean currentMode;
		int currentLine = m_startLine + m_lineCount;
		Unit aUnit;

		currentMode = true;
		currentBlock = new Vector();
		m_totalLineCount = m_lineCount;

		int[] bPos = new int[]{bIdx};
		int[] cLine = new int[]{currentLine};
		while((aUnit = Unit.next(m_owner, buf, bPos, m_fileName, cLine)) != null)
		{
			bIdx = bPos[0];
			currentLine = cLine[0];
			if(aUnit instanceof SubDirective)
			{
				if(aUnit instanceof ElifDirective)
				{
					// A new directive starts here. The SubDirective is tossed.
					// When this directive encounters an #endif, we are done
					// also.
					//
					if(currentMode)
					{
						// We treat the #elif as if it where an #else followed
						// by
						// an #if. We won't find the #endif though, since it's
						// eaten
						// by the #if. So, in other words, when the #if is done,
						// we
						// are done.
						//
						int top = currentBlock.size();
						m_trueBlock = new Unit[top];
						currentBlock.copyInto(m_trueBlock);
						aUnit = new IfDirective(m_owner, m_fileName,
								currentLine);
						bIdx = aUnit.parse(buf, bIdx);
						m_falseBlock = new Unit[]{aUnit};
						m_totalLineCount += aUnit.getLineCount();
					}
					else
						this.parseError("#elif in #else in #if " + m_contents);
					return bIdx;
				}

				if(aUnit instanceof ElseDirective)
				{
					if(currentMode)
					{
						// Switch to false
						//
						int top = currentBlock.size();
						m_trueBlock = new Unit[top];
						currentBlock.copyInto(m_trueBlock);
						currentBlock = new Vector();
						currentMode = false;
						m_totalLineCount++;
					}
					else
						this.parseError("#else in #else in #if " + m_contents);
				}
				else
				{
					// Endif, we are done now.
					//
					int top = currentBlock.size();
					Unit[] v = new Unit[top];
					currentBlock.copyInto(v);
					if(currentMode)
						m_trueBlock = v;
					else
						m_falseBlock = v;
					m_totalLineCount++;
					return bIdx;
				}
				continue;
			}
			m_totalLineCount += aUnit.getLineCount();
			currentBlock.addElement(aUnit);
		}

		this.parseError("unterminated in #if " + m_contents);
		return bIdx;
	}

	public int parseSimple(StringBuffer bld, char[] buf, int bIdx)
	throws ParseException
	{
		char c;
		int[] lcnt = new int[1];
		int bTop = buf.length;

		bld.append("defined(");
		bIdx = skipWhite(buf, bIdx, lcnt);
		m_lineCount += lcnt[0];

		while(bIdx < bTop)
		{
			c = buf[bIdx];
			if(!(Character.isLetterOrDigit(c) || c == '_'))
				break;
			bld.append(c);
			++bIdx;
		}
		bld.append(')');
		bIdx = skipToStartOfLine(buf, bIdx, lcnt);
		m_lineCount += lcnt[0];
		this.parseExpression(bld.toString().toCharArray(), 0);
		return this.parseContents(buf, bIdx);
	}

	private Unit[]		m_trueBlock;

	private Unit[]		m_falseBlock;

	private Expression	m_expression;

	private int			m_totalLineCount;
}