package se.tada.preproc;

import java.io.CharArrayWriter;
import java.io.IOException;
import java.io.Writer;

class PlainText extends Unit
{
	public PlainText(PreProcessor owner, String fileName, int startLine)
	{
		super(owner, fileName, startLine);
		m_significant = false;
	}

	public void expand()
	throws ParseException
	{
		// One reExpand is enough on top level. Each macro will
		// expand into a PlainText instance which in turn will
		// use the expandTo method to resolve marcos within
		// marcros.
		//
		try
		{
			this.reExpand(m_owner, true);
		}
		catch(IOException e)
		{
			throw new ParseException(e);
		}
	}

	// Expand over and over again until no more macro expansions
	// are made.
	//
	public void expandTo(Writer here) throws IOException
	{
		// TODO: Detect recursive macro expansions.
		//
		String tmp = m_contents;
		CharArrayWriter wrt = new CharArrayWriter(tmp.length() + 30);
		while(this.reExpand(wrt, false))
		{
			m_contents = wrt.toString();
			wrt.reset();
		}
		wrt.writeTo(here);
		m_contents = tmp;
	}

	public boolean isSignificant()
	{
		return m_significant;
	}

	public int parse(char[] buf, int bIdx)
	{
		char c;
		boolean inQuote = false;
		int[] lcnt = new int[]{0};
		int start = bIdx;
		int oIdx = bIdx;
		int bTop = buf.length;
		boolean signif = false;

		for(;;)
		{
			if(bIdx >= bTop)
				break;

			switch(c = buf[bIdx++])
			{
				case ESCNL_MARKER:
					// Backslash newlines means that the line should
					// be concatenated. Macros matching either side will
					// not be recognized later. Macros matching something
					// that is the result of the concatenation will be
					// recognized.
					//
					m_lineCount++;
					if(inQuote)
						//
						// Backslash newline within qoutes are not
						// removed.
						//
						buf[oIdx++] = c;
					continue;

				case '\\':
					signif = true;
					buf[oIdx++] = c;
					if(bIdx < bTop)
						buf[oIdx++] = buf[bIdx++];
					continue;

				case '"':
					inQuote = !inQuote;
					signif = true;
					buf[oIdx++] = c;
					continue;

				case '/':
					if(inQuote || bIdx == bTop)
					{
						signif = true;
						buf[oIdx++] = c;
						continue;
					}

					c = buf[bIdx++];
					if(c == '/')
					{
						// Start of CPPComment. Just skip to the next
						// line but keep the newline.
						//
						bIdx = skipToStartOfLine(buf, bIdx, lcnt);
						int cnt = lcnt[0];
						--cnt;
						--bIdx;
						m_lineCount += cnt;
						while(cnt > 0)
						{
							buf[oIdx++] = '\n';
							--cnt;
						}
						continue;
					}

					if(c == '*')
					{
						// Start of CComment. This is a bit tricky. Assume
						// that we have something like "FOO/* a comment */BAR"
						// The result should be FOOBAR. However if a macro is
						// defined for FOO or BAR, the macro should be expanded.
						// A macro defined for FOOBAR should NOT be expanded.
						//
						boolean[] ok = new boolean[]{true};
						bIdx = skipCComment(buf, bIdx, lcnt, ok);
						int cnt = lcnt[0];
						if(!ok[0])
							this.parseError("unterminated comment");
						m_lineCount += cnt;
						if(cnt > 0)
						{
							while(cnt > 0)
							{
								buf[oIdx++] = '\n';
								--cnt;
							}
						}
						else
						{
							if(bIdx < bTop)
							{
								c = buf[bIdx];
								if(Character.isLetter(c) || c == '_')
									buf[oIdx++] = (char)NULL_MARKER;
							}
						}
						continue;
					}
					signif = true;
					buf[oIdx++] = '/';
					buf[oIdx++] = c;
					if(inQuote && c == '"')
						inQuote = false;
					continue;

				case '\n':
					m_lineCount++;
					buf[oIdx++] = c;
					if(!inQuote)
					{
						if(bIdx < bTop && buf[bIdx] == '#')
							//
							// Start of next Unit (Directive)
							//
							break;
					}
					continue;

				case ' ':
				case '\t':
				case '\f':
					buf[oIdx++] = c;
					continue;

				default:
					signif = true;
					buf[oIdx++] = c;
					continue;
			}
			break;
		}

		if(signif == true)
		{
			if(m_owner.doingDependencyCheck())
				signif = false;
			else
				m_contents = String.valueOf(buf, start, oIdx - start);
		}
		m_significant = signif;
		return bIdx;
	}

	protected boolean reExpand(Writer out, boolean finalExp)
	throws IOException
	{
		char c;
		boolean inQuote = false;
		boolean reExp = false;
		char[] buf = m_contents.toCharArray();
		int bTop = buf.length;
		int bIdx = 0;
		int lIdx = 0;

		while(bIdx < bTop)
		{
			c = buf[bIdx++];
			if(inQuote)
			{
				// We don't recognize macros within quotes
				//
				if(c == '\\')
				{
					if(bIdx < bTop)
						++bIdx;
					continue;
				}

				if(c == '"')
					inQuote = false;
				continue;
			}

			switch(c)
			{
				case NULL_MARKER:
				case ESCNL_MARKER:
				case '\n':
					break;

				case '\\':
					// Treat next character literally.
					//
					if(bIdx < bTop)
						++bIdx;
					break;

				case '\'':
					bIdx = skipSingleQuote(buf, bIdx);
					break;

				case '"':
					inQuote = true;
					break;

				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':

					// Numbers must be skipped in order to avoid
					// attempt to do macro substitusions on things
					// line 012ADAM
					//
					while(bIdx < bTop)
					{
						c = buf[bIdx];
						if(!(Character.isLetterOrDigit(c) || c == '_'))
							break;
						++bIdx;
					}
					break;

				default:
					if(!Character.isLetter(c))
						break;

				// Fall through to '_':
				//
				case '_':
				// Possble macro substitution
				//
				{
					int[] tmp = new int[]{--bIdx};
					String id = getIdentifier(buf, tmp);

					DefineDirective macro = m_owner.getMacro(id);
					if(macro == null)
					{
						bIdx = tmp[0];
						break;
					}

					// Print all up to macro expansion start
					//
					if(bIdx > lIdx)
						out.write(buf, lIdx, bIdx - lIdx);

					// Expand macro and start next print after
					// that.
					//
					bIdx = tmp[0];
					tmp[0] = 0;
					bIdx = macro.expandMacro(out, this, buf, bIdx, tmp);
					lIdx = bIdx;
					reExp = true;
					if(finalExp)
					{
						// The parameters to a macro may be separated with
						// optional whitespace. This whitespace may be a
						// newline. Such newlines must then be expanded
						// after the macro.
						//
						while(tmp[0] > 0)
						{
							out.write('\n');
							tmp[0]--;
						}
					}
				}
					break;
			}
		}

		if(bIdx > lIdx)
			out.write(buf, lIdx, bIdx - lIdx);
		return reExp;
	}

	private boolean	m_significant;
}