package se.tada.preproc;

import java.io.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.Stack;

import se.tada.util.PathUtils;
import se.tada.util.StringUtils;
import se.tada.util.eval.Expression;
import se.tada.util.eval.Parser;

public class PreProcessor extends Writer implements MarkerTypes
{

	private final HashMap		m_bufferCache = new HashMap();

	private final HashMap		m_resolveCache = new HashMap();

	private final HashMap		m_macros = new HashMap();

	private final ArrayList		m_errors = new ArrayList();

	private final Stack			m_fileStack = new Stack();
	
	private Parser				m_parser;

	private Writer				m_result;

	private HashMap				m_initialMacros;

	private Set					m_dependencies;

	private String[]			m_includeDirs;

	private String				m_currentFileName;

	private int					m_currentLine;

	private int					m_lastLine;

	private boolean				m_lineSetPending;

	private int					m_writeCount;

	private int					m_atBOL;

	public PreProcessor()
	{
		m_result = null;
		m_initialMacros = null;
		m_dependencies = null;
		m_includeDirs = new String[0];
		m_currentFileName = "";
		m_currentLine = 0;
		m_lastLine = Integer.MIN_VALUE;
		m_lineSetPending = false;
		m_writeCount = 0;
		m_atBOL = 0;
	}

	public void close() throws IOException
	{
		m_result.close();
	}

	public Expression compile(String aString)
	throws ParseException
	{
		StringReader rdr = new StringReader(aString);
		if(m_parser == null)
			m_parser = new Parser(rdr);
		else
			m_parser.ReInit(rdr);
		try
		{
			return m_parser.expression();
		}
		catch(se.tada.util.eval.ParseException e)
		{
			throw new ParseException(e);
		}
	}

	public void flush() throws IOException
	{
		m_result.flush();
	}

	public void write(char[] buf, int start, int count) throws IOException
	{
		if(m_dependencies != null)
			return;

		int bIdx = start;
		int bTop = bIdx + count;
		Writer o = m_result;

		while(bIdx < bTop)
		{
			int rIdx;
			char c = buf[bIdx];
			switch(c)
			{
				case SOURCE_PUSH:
					if(bIdx > start)
					{
						o.write(buf, start, bIdx - start);
						m_writeCount += bIdx - start;
					}
					start = ++bIdx;
					while(buf[bIdx] != (char)SOURCE_PUSH)
						++bIdx;
					String file = String.valueOf(buf, start, bIdx - start);
					start = ++bIdx;
					while(buf[bIdx] != (char)SOURCE_PUSH)
						++bIdx;
					rIdx = Integer.valueOf(
							String.valueOf(buf, start, bIdx - start))
							.intValue();
					start = ++bIdx;
					this.pushFile(file, rIdx);
					continue;

				case SOURCE_POP:
					if(bIdx > start)
					{
						o.write(buf, start, bIdx - start);
						m_writeCount += bIdx - start;
					}
					start = ++bIdx;
					this.popFile();
					continue;

				case ESCNL_MARKER:
					if(bIdx > start)
					{
						o.write(buf, start, bIdx - start);
						m_writeCount += bIdx - start;
					}
					start = ++bIdx;
					o.write("\\\n");
					m_writeCount += 2;
					++m_currentLine;
					continue;

				case '\n':
					if(bIdx > start)
					{
						o.write(buf, start, bIdx - start);
						m_writeCount += bIdx - start;
					}
					while(bIdx < bTop && buf[bIdx] == '\n')
					{
						++m_currentLine;
						++bIdx;
					}
					start = bIdx;
					m_lineSetPending = true;
					continue;

				case NULL_MARKER:
				case ' ':
				case '\t':
					// Treat lines with only whitespace as nothing
					//
					rIdx = bIdx + 1;
					while(rIdx < bTop)
					{
						c = buf[rIdx];
						if(!(c == ' ' || c == '\t' || c == NULL_MARKER))
							break;
						++rIdx;
					}

					if(rIdx == bTop)
						break;

					if(c == '\n')
					{
						if(bIdx > start)
						{
							o.write(buf, start, bIdx - start);
							m_writeCount += bIdx - start;
						}
						start = rIdx;
						bIdx = rIdx; // Newline handled on next iteration.
						continue;
					}

					if(c == NULL_MARKER)
					{
						if(bIdx > start)
						{
							o.write(buf, start, bIdx - start);
							m_writeCount += bIdx - start;
						}
						start = ++bIdx;
						continue;
					}

				// Fall through to default
				//
				default:
					if(m_lineSetPending)
					{
						if(bIdx > start)
						{
							o.write(buf, start, bIdx - start);
							m_writeCount += bIdx - start;
							start = bIdx;
						}
						doLineSet();
					}
					++bIdx;
					continue;
			}
			break;
		}
		if(bIdx > start)
		{
			o.write(buf, start, bIdx - start);
			m_writeCount += bIdx - start;
		}
	}

	public void setLine(int aLine)
	{
		if(aLine == m_currentLine)
			return;
		m_currentLine = aLine;
		m_lineSetPending = true;
	}

	public void doLineSet() throws IOException
	{
		if(m_dependencies != null || m_lineSetPending == false)
			return;

		char[] nl = new char[]{'\n'};
		Writer o = m_result;
		m_lineSetPending = false;
		if(m_lastLine != Integer.MIN_VALUE)
		{
			int diff = m_currentLine - m_lastLine;
			if(diff < 10)
			{
				while(diff-- > 0)
				{
					o.write(nl);
					m_writeCount += nl.length;
				}
				m_lastLine = m_currentLine;
				m_atBOL = m_writeCount;
				return;
			}
		}
		if(m_writeCount != m_atBOL)
		{
			o.write(nl);
			m_writeCount += nl.length;
		}

		StringBuffer out = new StringBuffer(64);
		out.append("#line ");
		out.append(m_currentLine);
		out.append(" \"");
		out.append(m_currentFileName);
		out.append("\"");
		out.append(nl);
		String lnd = out.toString();
		o.write(lnd);
		m_writeCount += lnd.length();

		m_atBOL = m_writeCount;
		m_lastLine = m_currentLine;
	}

	public void pushFile(String aFileName, int lineNumber)
	{
		m_fileStack.push(new Object[]{m_currentFileName,
				new Integer(m_currentLine)});
		m_currentFileName = StringUtils.expandEscape(aFileName);
		m_currentLine = Integer.MIN_VALUE;
		m_lastLine = Integer.MIN_VALUE;
		this.setLine(lineNumber);
	}

	public void popFile()
	{
		Object[] entry = (Object[])m_fileStack.pop();
		if(entry == null)
			this.parseError(m_currentFileName, m_currentLine,
					"file stack underflow");

		m_currentFileName = (String)entry[0];
		m_currentLine = Integer.MIN_VALUE;
		m_lastLine = Integer.MIN_VALUE;
		this.setLine(((Integer)entry[1]).intValue());
	}

	public Unit[] parseFile(File aFile)
	throws ParseException
	{
		FileReader rdr = null;
		try
		{
			rdr = new FileReader(aFile);
			return this.parseStream(aFile.getAbsolutePath(), rdr);
		}
		catch(IOException e)
		{
			throw new ParseException(e);
		}
		finally
		{
			if(rdr != null)
				try { rdr.close(); } catch(IOException e) {}
		}
	}

	public Unit[] parseStream(String aFileName, Reader aStream)
	throws IOException, ParseException
	{
		ArrayList answer = new ArrayList();
		char[] tBuf = new char[4096];
		StringBuffer wBuf = new StringBuffer();
		int cnt;

		while((cnt = aStream.read(tBuf)) > 0)
			wBuf.append(tBuf, 0, cnt);

		String contents = wBuf.toString();
		char[] buf = contents.toCharArray();
		int bIdx = 0;
		int bTop = buf.length;
		int oIdx = 0;
		char c;
		Unit unit;

		// Convert the following (ESC == backslash) so that we can
		// simplify things a bit later on.
		//
		// CR NL -> NL
		// ESC NL -> ESCNL_MARKER
		// ESC CR NL -> ESCNL_MARKER
		//
		while(bIdx < bTop)
		{
			c = buf[bIdx++];
			switch(c)
			{
				case '\\':
					if(bIdx >= bTop)
						break;
					c = buf[bIdx++];
					if(c == '\n')
						c = ESCNL_MARKER;
					else if(c == '\r' && bIdx < bTop && buf[bIdx] == '\n')
					{
						c = ESCNL_MARKER;
						bIdx++;
					}
					else
						buf[oIdx++] = '\\';
					break;

				case '\r':
					if(bIdx < bTop && buf[bIdx] == '\n')
					{
						c = '\n';
						++bIdx;
					}
			}
			buf[oIdx++] = c;
		}

		// Truncate buffer if necessary.
		//
		if(oIdx < bTop)
		{
			char[] nBuf = new char[oIdx];
			System.arraycopy(buf, 0, nBuf, 0, oIdx);
			buf = nBuf;
		}

		int[] bPos = new int[]{0};
		int[] lnp = new int[]{0};
		while((unit = Unit.next(this, buf, bPos, aFileName, lnp)) != null)
			answer.add(unit);

		return (Unit[])answer.toArray(new Unit[answer.size()]);
	}

	public Unit[] resolveInclude(IncludeDirective aDirective, String[] pf)
	throws ParseException
	{
		File vf = this.resolveName(aDirective);
		if(vf == null)
		{
			// Will only happend during dependency check. We
			// silently ignore this.
			//
			pf[0] = aDirective.getFileName();
			return null;
		}

		String path;
		try
		{
			path = vf.getCanonicalPath();
		}
		catch(IOException e)
		{
			throw new ParseException(e);
		}

		Unit[] v = (Unit[])m_bufferCache.get(path);
		if(v == null)
		{
			v = this.parseFile(vf);
			m_bufferCache.put(path, v);
		}
		pf[0] = path;
		return v;
	}

	public void addDependency(String path)
	{
		if(m_dependencies != null)
			m_dependencies.add(path);
	}

	public void defineMacro(DefineDirective aMacro)
	{
		m_macros.put(aMacro.name(), aMacro);
	}

	public boolean doingDependencyCheck()
	{
		return m_dependencies != null;
	}

	public DefineDirective getMacro(String aMacroName)
	{
		return (DefineDirective)m_macros.get(aMacroName);
	}

	public Map getMacros()
	{
		return m_macros;
	}

	public void parseError(String aFile, int aLine, String aMessage)
	{
		StringBuffer bld = new StringBuffer(64);
		bld.append(aFile);
		bld.append(" : ");
		bld.append(aLine);
		bld.append(" : ");
		bld.append(aMessage);
		bld.append("\n");
		m_errors.add(bld.toString());
	}

	public String[] getDependencies(String inPath, Reader in)
	throws IOException, ParseException
	{
		m_errors.clear();
		m_macros.clear();
		m_macros.putAll(m_initialMacros);
		m_dependencies = new TreeSet();

		Unit[] units = this.parseStream(inPath, in);
		int top = units.length;

		m_result = null;
		this.pushFile(inPath, 1);
		for(int idx = 0; idx < top; ++idx)
			units[idx].expand();

		String[] answer = new String[m_dependencies.size()];
		m_dependencies.toArray(answer);
		return answer;
	}

	public void process(String inPath, Reader in, Writer out)
	throws ParseException
	{
		m_errors.clear();
		m_macros.clear();
		m_macros.putAll(m_initialMacros);

		try
		{
			Unit[] units = this.parseStream(inPath, in);
			int top = units.length;
			m_result = out;
			this.pushFile(inPath, 1);
			for(int idx = 0; idx < top; ++idx)
				units[idx].expand();
		}
		catch(IOException e)
		{
			throw new ParseException(e);
		}

		if(m_errors.size() > 0)
			throw new ParseException(m_errors);
	}

	public File resolveName(IncludeDirective aDirective)
	{
		String aName = aDirective.getContents();
		File resolved = (File)m_resolveCache.get(aName);
		if(resolved != null)
			return resolved;

		resolved = PathUtils.fileInPath(m_includeDirs, aName);
		if(resolved != null)
		{
			m_resolveCache.put(aName, resolved);
			return resolved;
		}
		return null;
	}

	public void undefineMacro(UndefDirective aMacro)
	{
		m_macros.put(aMacro.name(), null);
	}

	public static void main(String[] argv)
	{
		try
		{
			PreProcessor pp = PreProcessor.on(new String[]{"."}, new String[] {});
			Writer out = new OutputStreamWriter(System.out);
			Reader in = new FileReader(argv[0]);
			pp.process(argv[0], in, out);
			out.flush();
		}
		catch(Exception e)
		{
			System.out.println(e.toString());
		}
	}

	public static PreProcessor on(String[] includes, String[] defines)
	{
		PreProcessor answer = new PreProcessor();
		answer.m_includeDirs = includes;

		int top = defines.length;
		HashMap macros = new HashMap();
		for(int idx = 0; idx < top; ++idx)
		{
			String define = defines[idx];
			String value;
			int splitPos = define.indexOf('=');
			if(splitPos >= 0)
			{
				value = define.substring(splitPos + 1);
				define = define.substring(0, splitPos);
			}
			else
				value = "1";

			DefineDirective dd = new DefineDirective(answer, "", 0);
			dd.parse((define + " " + value).toCharArray(), 0);
			macros.put(define, dd);
		}

		answer.m_initialMacros = macros;
		return answer;
	}
}
