package se.tada.util.printf;

public abstract class FloatFormat extends NumberFormat
{
	private final boolean m_sharp;

	private final char m_expChar;

	public FloatFormat(int prefixType, boolean sharp, boolean uppercase)
	{
		super(prefixType, 'l');
		m_sharp = sharp;
		m_expChar = uppercase ? 'E' : 'e';
	}

	protected String basicFormat(Object src, int width, int precision)
	{
		double val = (src instanceof Number) ? ((Number)src).doubleValue()
				: Double.valueOf(src.toString()).doubleValue();

		if(Double.isInfinite(val))
			return val < 0.0 ? "-Inf" : "Inf";

		if(Double.isNaN(val))
			return "NaN";

		if(precision < 0)
			precision = 6;
		else if(precision == 0 && m_sharp)
			precision = 1;

		int[] tmp = new int[1];
		long mantissa = getDecimalMantissaAndExponent(val, tmp);
		String str = this.makeString(mantissa, tmp[0], precision);

		// If the number is about to be zero padded, see to that
		// a minus sign is moved to the left of the pad.
		//
		int len = str.length();
		if(width > len && this.isZeroPad() && val < 0.0)
		{
			int pLen = 1 + width - len;
			int pIdx = 0;
			char[] pad = new char[width];

			pad[pIdx++] = '-';
			while(pIdx < pLen)
				pad[pIdx++] = '0';
			str.getChars(1, len, pad, pIdx);
			str = String.valueOf(pad, 0, width);
		}
		return str;
	}

	public static long getDecimalMantissaAndExponent(double val, int[] expRet)
	{
		String str = Double.toString(Math.abs(val));
		int len = str.length();
		char[] buf = new char[40];
		int idx = str.indexOf('E');

		int exponent = 0;
		long mantissa = 0L;

		if(idx >= 0)
		{
			// The number will be in the form m.mmmEseee where
			// m.mmm is the mantissa, E is the character 'E', s
			// is the sign (+ or -) and eee is the exponent.
			//
			// First we remove the decimal point
			//
			buf[0] = str.charAt(0);
			str.getChars(2, len, buf, 1);
			--len;
			--idx;

			// Parse the exponent.
			//
			exponent = Integer.parseInt(String.valueOf(buf, idx + 1, len
					- (idx + 1)));

			// Find last nonzero digit
			//
			for(int r = idx - 1; r >= 0; --r)
				if(buf[r] != '0')
				{
					idx = r + 1;
					break;
				}
			mantissa = Long.parseLong(String.valueOf(buf, 0, idx));
		}
		else
		{
			int dp = str.indexOf('.');
			str.getChars(0, dp, buf, 0);
			str.getChars(dp + 1, len, buf, dp);
			--len;

			// Find first nonzero digit
			//
			int f = -1;
			for(idx = 0; idx < len; ++idx)
				if(buf[idx] != '0')
				{
					f = idx;
					break;
				}

			if(f >= 0)
			{
				exponent = dp - f - 1;

				// Find last nonzero digit
				//
				int r = len;
				for(idx = len - 1; idx >= f; --idx)
					if(buf[idx] != '0')
					{
						r = idx + 1;
						break;
					}

				mantissa = Long.parseLong(String.valueOf(buf, f, r - f));
			}
		}

		if(val < 0.0)
			mantissa = -mantissa;

		expRet[0] = exponent;
		return mantissa;
	}

	protected int isRoundUp(char[] buf, int idx, int top)
	{
		if(buf[0] < '5')
			return 0;
		if(buf[0] > '5')
			return 1;

		for(; idx < top; ++idx)
			if(buf[idx] != '0')
				return 1;

		return 2;
	}

	/**
	 * 
	 * @param mantissa
	 *            The actual value to be formatted with respect to the
	 *            <i>exponent </i>
	 * @param exponent
	 *            The exponent of the <i>mantissa </i>
	 * @param precision
	 *            Either the maximum number of significant digits or the number
	 *            of digits after the decimal point depending on <i>compact </i>
	 * @param compact
	 *            when <c>true </c>, the <i>precision </i> denotes the number of
	 *            significant digits printed as opposed to the number of digits
	 *            after the decimal point. <code>true</code> also stipulates
	 *            that no trailing zeros should be added.
	 * @return The formatted <c>String </c>
	 */

	protected String makeDecString(long mantissa, int exponent, int precision,
			boolean compact)
	{
		char[] buf = new char[40 + Math.abs(exponent)];
		int bIdx = 0;
		int i;

		if(mantissa < 0)
		{
			buf[bIdx++] = '-';
			mantissa = -mantissa;
		}

		String str = Long.toString(mantissa, 10);
		int len = str.length();

		int fracStart = -1;
		if(exponent >= len - 1)
		{
			// All significant digits are to the left of the
			// decimal point.
			//
			str.getChars(0, len, buf, bIdx);
			bIdx += len;
			int t = exponent - len + 1;
			for(i = 0; i < t; i++)
				buf[bIdx++] = '0';

			if(compact && !m_sharp)
				//
				// Don't add any trailing zeroes.
				//
				precision = 0;
			else if(precision > 0)
			{
				buf[bIdx++] = '.';
				fracStart = bIdx;
			}
		}
		else if(exponent < 0)
		{
			// All significant digits are to the right of the
			// decimal point.
			//
			if(precision > 0)
			{
				buf[bIdx++] = '0';
				buf[bIdx++] = '.';
				fracStart = bIdx;
				for(i = exponent; i < -1; i++)
					buf[bIdx++] = '0';
				str.getChars(0, len, buf, bIdx);
				bIdx += len;

				if(compact)
					precision -= 1;
			}
			else
			{
				if(exponent == -1 && isRoundUp(str.toCharArray(), 0, len) == 1)
					//
					// Number is greater than 0.5
					//
					buf[bIdx++] = '1';
				else
					//
					// Number is exactly 0.5 or less.
					//
					buf[bIdx++] = '0';
			}
		}
		else
		{
			str.getChars(0, exponent + 1, buf, bIdx);
			bIdx += exponent + 1;

			if(compact)
				precision -= exponent + 1;

			if(precision > 0)
			{
				buf[bIdx++] = '.';
				fracStart = bIdx;
				str.getChars(exponent + 1, len - (exponent + 1), buf, bIdx);
				bIdx += len - (exponent + 1);
			}
			else
			{
				int roundPos = bIdx - 1;
				switch(isRoundUp(str.toCharArray(), exponent + 1, len))
				{
					case 0: // Don't round up. Fraction is less than 0.5
						break;
					case 1: // Round up. Fraction is greater than 0.5
						buf[roundPos]++;
						break;
					default: // Round up if uneven. Fraction is exactly 0.5
						if((Character.digit(buf[roundPos], 10) & 1) != 0)
							buf[roundPos]++;
				}
			}
		}

		if(fracStart > 0)
		{
			int fracLen = bIdx - fracStart;
			if(fracLen > precision)
			{
				// Fraction contains to many digits. We must round it
				//
				int roundPos = fracStart + precision - 1;
				switch(isRoundUp(buf, roundPos + 1, fracLen))
				{
					case 0: // Don't round up.
						break;
					case 1: // Round up.
						buf[roundPos]++;
						break;
					default: // Round up if uneven
						if((Character.digit(buf[roundPos], 10) & 1) != 0)
							buf[roundPos]++;
				}
				bIdx = ++roundPos;
			}
			else
			{
				if(m_sharp || !compact)
					for(i = fracLen; i < precision; ++i)
						buf[bIdx++] = '0';
			}
		}
		return String.valueOf(buf, 0, bIdx);
	}

	/**
	 * 
	 * @param mantissa
	 *            The actual value to be formatted with respect to the
	 *            <i>exponent </i>
	 * @param exponent
	 *            The exponent of the <i>mantissa </i>
	 * @param precision
	 *            Either the maximum number of significant digits or the number
	 *            of digits after the decimal point depending on <i>compact </i>
	 * @param compact
	 *            when <c>true </c>, the <i>precision </i> denotes the number of
	 *            significant digits printed as opposed to the number of digits
	 *            after the decimal point. <code>true</code> also stipulates
	 *            that no trailing zeros should be added.
	 * @return The formatted <c>String </c>
	 */

	protected String makeExpString(long mantissa, int exponent, int precision,
			boolean compact)
	{
		char[] buf = new char[40];
		int bIdx = 0;

		if(mantissa < 0)
		{
			buf[bIdx++] = '-';
			mantissa = -mantissa;
		}

		// If not compact, the precision denotes the number of digits
		// after the decimal point. Using exponential notation we will
		// have exactly one digit before the decimal point. Increasing
		// precision with one will yield the number of significant
		// digits.
		//
		if(!compact || precision == 0)
			precision++;

		String str = Long.toString(mantissa, 10);
		int len = str.length();

		if(len > precision)
		{
			// Pick the part to be rounded and determine
			// what the first character should be rounded
			// into.
			//
			char rc = str.charAt(precision - 1);
			switch(isRoundUp(str.toCharArray(), precision, len))
			{
				case 0:
					break;
				case 1:
					rc++;
					break;
				default:
					if((Character.digit(rc, 10) & 1) != 0)
						rc++;
			}

			if(precision > 1)
			{
				buf[bIdx++] = str.charAt(0);
				buf[bIdx++] = '.';
				str.getChars(1, precision - 1, buf, bIdx);
				bIdx += precision - 2;
				buf[bIdx++] = rc;
			}
			else
				buf[bIdx++] = rc;
		}
		else if(precision > len)
		{
			buf[bIdx++] = str.charAt(0);
			if(len > 1 || m_sharp || !compact)
			{
				buf[bIdx++] = '.';
				if(len > 1)
				{
					str.getChars(1, len, buf, bIdx);
					bIdx += len - 1;
				}
				else
				{
					buf[bIdx++] = '0';
					++len;
				}

				if(m_sharp || !compact)
				{
					while(len < precision)
					{
						buf[bIdx++] = '0';
						++len;
					}
				}
			}
		}
		else
		{
			buf[bIdx++] = str.charAt(0);
			if(len > 1)
			{
				buf[bIdx++] = '.';
				str.getChars(1, len, buf, bIdx);
				bIdx += len - 1;
			}
		}

		buf[bIdx++] = m_expChar;
		buf[bIdx++] = (exponent >= 0) ? '+' : '-';
		str = Integer.toString(Math.abs(exponent));
		exponent = str.length();

		// Exponent is always printed with at least 2 digits
		//
		if(exponent == 1)
			buf[bIdx++] = '0';

		str.getChars(0, exponent, buf, bIdx);
		bIdx += exponent;
		return String.valueOf(buf, 0, bIdx);
	}

	protected abstract String makeString(long mantissa, int exponent,
			int precision);
}