package se.tada.util.printf;

import java.io.IOException;
import java.io.StringWriter;
import java.io.Writer;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.text.FieldPosition;
import java.text.Format;
import java.text.ParsePosition;
import java.util.ArrayList;
import java.util.HashMap;

import se.tada.util.StringUtils;

public class PrintFormat extends Format implements PrefixType
{
	private static final long serialVersionUID = -5180435320434995880L;

	static private final HashMap s_cache = new HashMap();

	private final Argument[] m_arguments;

	private final Part[] m_parts;

	public static String va_form(String str, Object[] args)
	{
		try
		{
			Writer w = new StringWriter();
			PrintFormat.forPattern(str).format(args, w);
			return w.toString();
		}
		catch(IOException e)
		{
			throw new RuntimeException(e);
		}
	}

	private class Argument
	{
		private final int m_number;

		Argument(int number)
		{
			m_number = number;
		}

		Object get(Object[] arglist)
		{
			return m_number < arglist.length ? arglist[m_number] : null;
		}
	}

	private class MessageArgument extends Argument
	{
		private final String m_message;

		MessageArgument(String message, int number)
		{
			super(number);
			m_message = message;
		}

		Object get(Object[] arglist)
		{
			Object base = super.get(arglist);
			if(base == null)
				return null;

			boolean isClass = (base instanceof Class);
			Class cls = isClass ? (Class)base : base.getClass();

			try
			{
				try
				{
					java.lang.reflect.Field fld = cls.getField(m_message);
					if(isClass)
					{
						if(!Modifier.isStatic(fld.getModifiers()))
							throw new NoSuchFieldException("static " + m_message);
						return fld.get(null);
					}
					return fld.get(base);
				}
				catch(NoSuchFieldException e)
				{
					Method gtr = cls.getMethod(StringUtils.makeGetter(m_message), null);
					if(isClass)
					{
						if(!Modifier.isStatic(gtr.getModifiers()))
							throw new NoSuchMethodException("static " + m_message);
						return gtr.invoke(null, null);
					}
					return gtr.invoke(base, null);
				}
			}
			catch(NoSuchMethodException e)
			{
				return null;
			}
			catch(InvocationTargetException e)
			{
				Throwable t = e.getTargetException();
				if(t instanceof RuntimeException)
					throw (RuntimeException)t;
				return null;
			}
			catch(IllegalAccessException e)
			{
				return null;
			}
		}
	}

	private class CompoundArgument extends Argument
	{
		CompoundArgument(Argument[] array)
		{
			super(0);
			m_array = array;
		}

		Object get(Object[] arglist)
		{
			int top = m_array.length;
			Object[] result = new Object[top];
			for(int idx = 0; idx < top; ++idx)
				result[idx] = m_array[idx].get(arglist);
			return result;
		}

		private Argument[] m_array;
	}

	static public PrintFormat forPattern(String aPattern)
	{
		PrintFormat v = (PrintFormat)s_cache.get(aPattern);
		if(v == null)
		{
			v = new PrintFormat(aPattern);
			s_cache.put(aPattern, v);
		}
		return v;
	}

	public PrintFormat(String pattern)
	{
		Argument[] av = null;
		ArrayList parts = new ArrayList();
		char[] format = pattern.toCharArray();
		int fTop = format.length;
		int fIdx = 0;

		// Check if the format string starts with an argument reorder
		// statement in the form "%(n n ...) "
		//
		if(format[fIdx] == '%' && format[fIdx + 1] == '(')
		{
			ArrayList a = new ArrayList();
			fIdx = parseReorder(format, fIdx + 2, a);
			av = (Argument[])a.toArray(new Argument[a.size()]);
		}

		for(;;)
		{
			// Find first format sequence. Starting with '%'.
			//
			boolean gotPercent = false;
			int bp = fIdx;
			while(fIdx < fTop)
			{
				if(format[fIdx++] == '%')
				{
					if(fIdx < fTop && format[fIdx] == '%')
					{
						gotPercent = true;
						++fIdx;
						continue;
					}
					--fIdx;
					break;
				}
			}

			if(fIdx > bp)
			{
				// ordinary non-% characters
				//
				String p = String.valueOf(format, bp, fIdx - bp);
				if(gotPercent)
				{
					// The string contains '%%' constructs.
					// They must be transformed into '%'
					//
					char[] o = p.toCharArray();
					int t = o.length;
					int ni = 0;
					for(int oi = 0; oi < t;)
						if((o[ni++] = o[oi++]) == '%')
							oi++;
					p = String.valueOf(o, 0, ni);
				}
				parts.add(new SimpleTextFormat(p));
			}

			if(fIdx >= fTop)
				//
				// end of format
				//
				break;

			fIdx = parseFormat(format, fIdx + 1, parts);
		}

		fTop = parts.size();
		m_parts = (Part[])parts.toArray(new Part[fTop]);

		// Calculate the expected number of arguments. Should
		// a part that uses more than one argument be found, then
		// remember this so that we can modify the argument vector
		// later.
		//
		boolean hasCompound = false;
		int aIdx = 0;

		for(fIdx = 0; fIdx < fTop; ++fIdx)
		{
			int c = m_parts[fIdx].getRequiredNumberOfArguments();
			if(c > 1)
				hasCompound = true;
			aIdx += c;
		}

		// Fix the argument handlers so that they correspond to the
		// actual number of arguments in the format string.
		//
		if(av == null || av.length < aIdx)
		{
			Argument[] newAv = new Argument[aIdx];

			if(av != null)
			{
				fIdx = av.length;
				System.arraycopy(av, 0, newAv, 0, fIdx);
			}
			else
				fIdx = 0;

			for(; fIdx < aIdx; ++fIdx)
				newAv[fIdx] = new Argument(fIdx);
			av = newAv;
		}

		if(hasCompound)
		{
			// Find the parts that requires compound arguments. Adjust
			// the argument vector accordingly.
			//
			aIdx = 0;
			for(fIdx = 0; fIdx < fTop; ++fIdx)
			{
				int c = m_parts[fIdx].getRequiredNumberOfArguments();
				if(c > 1)
				{
					int i;
					int t;
					Argument[] array = new Argument[c];
					for(i = 0; i < c; ++i)
						array[i] = av[aIdx + i];

					av[aIdx++] = new CompoundArgument(array);
					--c;

					i = aIdx;
					for(t = av.length - c; i < t; ++i)
						av[i] = av[i + c];
					t += c;
					for(; i < t; ++i)
						av[i] = null;
				}
				else if(c > 0)
					aIdx++;
			}
		}
		m_arguments = av;
	}

	private int parseFormat(char[] format, int fIdx, ArrayList vec)
	{
		// Flags - nonzero if corresponding character is in format
		//
		boolean leftAlign = false; // - left align output
		boolean sharp = false; // # flag (meaning vary depending on type)
		boolean zeroPad = false; // pad string with zeroes
		int prefix = NONE;
		Object width = null;
		Object prec = null;
		char c;

		for(;; ++fIdx)
		{
			// Scan the <flags>
			//
			switch(format[fIdx])
			{
				case '+':
					prefix = SIGN;
					continue;
				case '-':
					leftAlign = true;
					continue;
				case ' ':
					if(prefix == NONE)
						//
						// Ignored if '+' flag is given
						//
						prefix = SPACE;
					continue;
				case '#':
					sharp = true;
					continue;
			}
			break;
		}

		if(format[fIdx] == '0')
		{
			zeroPad = true;
			++fIdx;
		}

		// Scan the field width
		//
		if(format[fIdx] == '*')
		{
			// Width will be determined by argument
			//
			width = Boolean.TRUE;
			++fIdx;
		}
		else
		{
			int i = 0;
			for(; Character.isDigit(c = format[fIdx]); fIdx++)
				i = i * 10 + Character.getNumericValue(c);
			width = new Integer(i);
		}

		// Scan the precision
		//
		if(format[fIdx] == '.')
		{
			++fIdx;
			if(format[fIdx] == '*')
			{
				// Precision is supplied as an argument.
				//
				prec = Boolean.TRUE;
				++fIdx;
			}
			else
			{
				int i = 0;
				for(; Character.isDigit(c = format[fIdx]); fIdx++)
					i = i * 10 + Character.getNumericValue(c);
				prec = new Integer(i);
			}
		}

		// Scan the length modifier. It is meaningful
		// when unsigned quantities are printed. -32
		// will look radically different using %uh,
		// %u, or %ul
		//
		char size = 'i'; // Integer, the default
		switch(format[fIdx])
		{
			case 'l': // Long
			case 'h': // Short
			case 'b': // Byte
				size = format[fIdx++];
		}

		BaseFormat v;
		switch(c = format[fIdx++])
		{
			case 'd':
			case 'u':
				v = new DecimalIntFormat(prefix, size, c == 'u');
				break;
			case 'o':
				v = new OctalIntFormat(sharp ? OCTAL : NONE, size);
				break;
			case 'X':
				v = new HexIntFormat(sharp ? HEX : NONE, size, true);
				break;
			case 'x':
				v = new HexIntFormat(sharp ? HEX : NONE, size, false);
				break;
			case 'g':
				v = new GenericFloatFormat(prefix, sharp, false);
				break;
			case 'G':
				v = new GenericFloatFormat(prefix, sharp, true);
				break;
			case 'e':
				v = new ExponentialFloatFormat(prefix, sharp, false);
				break;
			case 'E':
				v = new ExponentialFloatFormat(prefix, sharp, true);
				break;
			case 'f':
				v = new DecimalFloatFormat(prefix, sharp);
				break;
			case 'c':
				v = new CharFormat();
				break;
			case 's':
				v = new TextFormat();
				break;
			default:
				throw new PrintFormatException(format, fIdx);
		}
		v.setWidth(width);
		v.setPrecision(prec);
		v.setLeftAligned(leftAlign);
		v.setZeroPad(zeroPad);
		vec.add(v);
		return fIdx;
	}

	public Object parseObject(String source, ParsePosition status)
	{
		// This method is derived from class java.text.Format
		// to do: code goes here
		return null;
	}

	public StringBuffer format(Object obj, StringBuffer out, FieldPosition pos)
	{
		StringWriter v = new StringWriter();
		try
		{
			this.format(obj, v);
		}
		catch(IOException e)
		{
		}
		out.append(v.toString());
		return out;
	}

	public void format(Object obj, Writer out) throws IOException
	{
		int aIdx = 0;
		int top = m_parts.length;
		Object[] args = (obj instanceof Object[]) ? (Object[])obj
				: new Object[] { obj };

		for(int idx = 0; idx < top; ++idx)
		{
			Part p = m_parts[idx];
			int n = p.getRequiredNumberOfArguments();
			if(n > 0)
			{
				p.format(m_arguments[aIdx].get(args), out);
				aIdx += n;
			}
			else
				p.format(null, out);
		}
	}

	private int parseReorder(char[] format, int fIdx, ArrayList vec)
	{
		// The reorder string may contain two types of entries,
		// an integer or a method call. The method call is in the
		// form 1->fooBar which means, call method fooBar on the
		// object found at first argument position.
		//
		int n;
		for(;;)
		{
			char c = format[fIdx];
			while(Character.isWhitespace(c))
				c = format[++fIdx];

			if(c == ')')
				//
				// End of format list. We're done with the reordering
				//
				break;

			// Convert the alpha numbers to an integer.
			//
			for(n = 0; Character.isDigit(c = format[++fIdx]); n = n * 10
					+ Character.getNumericValue(c))
				;

			if(n < 1)
				throw new PrintFormatException(format, fIdx);

			--n;
			if(c == '-' && format[fIdx + 1] == '>')
			{
				// This is a method call on object number 'n'.
				// in the form n->methodName
				//
				fIdx += 2;
				int symStart = fIdx;

				if(!Character.isJavaIdentifierStart(format[fIdx]))
					//
					// Symbol must start with alpha
					//
					throw new PrintFormatException(format, fIdx);

				// Calculate the boundaries of the symbol and
				// copy it into a new area.
				//
				while(Character.isJavaIdentifierPart(format[++fIdx]))
					;

				// Save current token after the symbol and replace
				// it with a '0'. Then a 'real' Symbol object can
				// be created from symStart.
				//
				String key = String.valueOf(format, symStart, fIdx - symStart);
				vec.add(new MessageArgument(key, n));
			}
			else
				vec.add(new Argument(n));
		}
		return fIdx;
	}
}